from django.db.models import Exists
from django.db.models import OuterRef
from django.db.models import Q

from kinder.core.children.models import ChildrenDelegate
from kinder.core.declaration_status.enum import DSS
from kinder.core.dict.models import DULDelegateType
from kinder.core.helpers import get_q
from kinder.core.helpers import recursive_py_getattr

# Соответствие полей представителя в ЭДС и полей из блока
# ApplicationOrderInfoRequest
DELEGATE_FIELDS_MAP = {
    'firstname': 'PersonInfo.PersonName',
    'surname': 'PersonInfo.PersonSurname',
    'dul_series': 'PersonIdentityDocInfo.IdentityDocSeries',
    'dul_number': 'PersonIdentityDocInfo.IdentityDocNumber',
    'dul_date': 'PersonIdentityDocInfo.IdentityDocIssueDate',
    'dul_issued_by': 'PersonIdentityDocInfo.IdentityDocIssued',
    'dul_place': 'PersonIdentityDocInfo.IdentityDocIssueCode',
    'dul_type': (
        'PersonIdentityDocInfo.IdentityDocName.code',
        lambda code: DULDelegateType.objects.filter(esnsi_code=code).first()
    )
}


def check_and_update_decls_delegates(declarations_query, request_body):
    """
    Проверяет существование представителей по заданным (необходимым для
    запроса OrderRequest) данным и если они не заполнены, пытается заполнить
    их данными из запроса ApplicationOrderInfoRequest.

    :param declarations_query: Заявки
    :type declarations_query: Queryset[Declaration]
    :param request_body: Тело запроса ApplicationOrderInfoRequest
    :type request_body: ApplicationOrderInfoRequest
    """

    exclude_q = (
        get_q('delegate__firstname', True) |
        get_q('delegate__surname', True) |
        get_q('delegate__dul_issued_by', True) |
        get_q('delegate__dul_place', True) |
        Q(delegate__dul_date__isnull=True) |
        Q(delegate__dul_type__isnull=True)
    )

    delegate_doc_info = request_body.PersonIdentityDocInfo
    child_delegate_dul_filter = Q(
        delegate__dul_number=delegate_doc_info.IdentityDocNumber)
    child_delegate_dul_filter &= (
       Q(delegate__dul_series=delegate_doc_info.IdentityDocSeries)
       if delegate_doc_info.IdentityDocSeries else
       get_q('delegate__dul_series', is_empty=True)
    )

    delegates_exists = Exists(
        ChildrenDelegate.objects.filter(
            child_delegate_dul_filter,
            children=OuterRef('children'),
        ).exclude(exclude_q)
    )

    delegates_declarations_query = declarations_query.annotate(
        delegates_exists=delegates_exists
    ).filter(delegates_exists=True).exclude(status__code=DSS.ARCHIVE)

    if not delegates_declarations_query.exists():
        for declaration in declarations_query.iterator():
            chdel_set = declaration.children.childrendelegate_set
            try:
                children_delegate = chdel_set.get(child_delegate_dul_filter)
            except ChildrenDelegate.DoesNotExist:
                continue

            delegate = children_delegate.delegate

            for delegate_field in DELEGATE_FIELDS_MAP.keys():
                if not getattr(delegate, delegate_field):
                    req_field = DELEGATE_FIELDS_MAP[delegate_field]

                    if isinstance(req_field, tuple):
                        req_field, req_function = req_field
                        req_value = recursive_py_getattr(
                            request_body, req_field)

                        req_value = (
                            req_function(req_value) if req_value else None
                        )
                    else:
                        req_value = recursive_py_getattr(
                            request_body, req_field)

                    setattr(delegate, delegate_field, req_value)

            delegate.save()

    return delegates_declarations_query
